% PT2PT_3D_COVARIANCE_CLOSEDFORM 
% Numerical solution for ICP covariance
%
%% Description:
%
% Author - 	
%           Sergi Foix Salmeron
%           PRE-JAE
%           Ernesto Homar Teniente Aviles
% 			CONACYT Fellow
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 10/2009
%
% Usage:
%   [ CovX ] = pt2pt_3D_Covariance( refPointCloud,
%   newPointCloud, xyzRxRyRz, CovZ )
%
% Variables:
%   Inputs:
%       refPointCloud   - a nx3 matrix, data set 1
%       newPointCloud   - a nx3 matrix, data set 2
%       xyzRxRyRz       -  a 1x6 vector, robot pose
%       CovZ            - a 3nx3n matrix, data set 1 points covariance
%
% 	Outputs:
%		CovX	- a 6x61 matrix, related pose covariance
%
%

%% Code
function CovX = pt2pt_3D_Covariance( refPointCloud, newPointCloud, xyzRxRyRz, CovZ )
% Estimacio de la transformacio
if size(refPointCloud)~=size(newPointCloud)
    error('Both input point clouds have to be of the same size');
end

tx      = xyzRxRyRz(1);
ty      = xyzRxRyRz(2);
tz      = xyzRxRyRz(3);
phi     = xyzRxRyRz(4);
theta   = xyzRxRyRz(5);
psi     = xyzRxRyRz(6);

Etot = 0;
Gtot = zeros(6,1);
EHess = zeros(6,6);

c.m = refPointCloud;
c.s = newPointCloud;

d2J_dxdxz = zeros(6,size(c.s,2));
d2J_dxdyz = zeros(6,size(c.s,2));
d2J_dxdzz = zeros(6,size(c.s,2));

% These are the interval to approximate the derivatives with the increments.
eps_xyz = 0.00001;
eps_rot = deg2rad(0.01);
a=1;
for ii=1:size(c.s,2)
    if rem(a,10)
        fprintf('* ');
    else
        fprintf('*\n');
    end
    a=a+1;
%     clc;
%     fprintf('%d\n',ii);
    
    x_z = c.s(1,ii);
    y_z = c.s(2,ii);
    z_z = c.s(3,ii);
    x_m = c.m(1,ii);
    y_m = c.m(2,ii);
    z_m = c.m(3,ii);

    % The error function is the traditional point-to-point distance (squared)
    E_k = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        sum((...
            applyHRt2point([ x_z_ y_z_ z_z_ ]', xyzRxRyRz2HRt([x_; y_; z_; phi_; theta_; psi_],'rad', 'pre')) ...
            - ...
            [ x_m y_m z_m ]' ...
            ).^2);
    
    %% First derivative
    gradEk = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        [ deriv(@(xx)E_k(x_z_, y_z_, z_z_, xx, y_, z_, phi_, theta_, psi_), x_, eps_xyz); ...
          deriv(@(yy)E_k(x_z_, y_z_, z_z_, x_, yy, z_, phi_, theta_, psi_), y_, eps_xyz); ...
          deriv(@(zz)E_k(x_z_, y_z_, z_z_, x_, y_, zz, phi_, theta_, psi_), z_, eps_xyz); ...
          deriv(@(ph)E_k(x_z_, y_z_, z_z_, x_, y_, z_, ph, theta_, psi_), phi_, eps_rot); ...
          deriv(@(tt)E_k(x_z_, y_z_, z_z_, x_, y_, z_, phi_, tt, psi_), theta_, eps_rot); ...
          deriv(@(ps)E_k(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, ps), psi_, eps_rot)];

    %% Second derivative
    d2Ek_dx2 = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        [ deriv(@(xx)gradEk(x_z_, y_z_, z_z_, xx, y_, z_, phi_, theta_, psi_), x_, eps_xyz) ...
          deriv(@(yy)gradEk(x_z_, y_z_, z_z_, x_, yy, z_, phi_, theta_, psi_), y_, eps_xyz) ...
          deriv(@(zz)gradEk(x_z_, y_z_, z_z_, x_, y_, zz, phi_, theta_, psi_), z_, eps_xyz) ...
          deriv(@(ph)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, ph, theta_, psi_), phi_, eps_rot) ...
          deriv(@(tt)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, phi_, tt, psi_), theta_, eps_rot) ...
          deriv(@(ps)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, ps), psi_, eps_rot)];

    %% The noise has three components: epsilon_i, epsilon_j1, epsilon_j2
    d2Ek_dxdexz = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        deriv(@(exz)gradEk(exz, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_), x_z_, eps_xyz);

    d2Ek_dxdeyz = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        deriv(@(eyz)gradEk(x_z_, eyz, z_z_, x_, y_, z_, phi_, theta_, psi_), y_z_, eps_xyz);

    d2Ek_dxdezz = @(x_z_, y_z_, z_z_, x_, y_, z_, phi_, theta_, psi_) ...
        deriv(@(ezz)gradEk(x_z_, y_z_, ezz, x_, y_, z_, phi_, theta_, psi_), z_z_, eps_xyz);

    Etot = Etot + E_k(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);
    Gtot = Gtot + gradEk(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);
    EHess = EHess + d2Ek_dx2(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);

    dxz_ii = d2Ek_dxdexz(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);
    dyz_ii = d2Ek_dxdeyz(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);
    dzz_ii = d2Ek_dxdezz(x_z, y_z, z_z, tx, ty, tz, phi, theta, psi);

    d2J_dxdxz(:,ii) = d2J_dxdxz(:,ii) + dxz_ii;
    d2J_dxdyz(:,ii) = d2J_dxdyz(:,ii) + dyz_ii;
    d2J_dxdzz(:,ii) = d2J_dxdzz(:,ii) + dzz_ii;

end

E2J_dxdz = [ d2J_dxdxz  d2J_dxdyz d2J_dxdzz ];

Jacobian = inv(EHess)*E2J_dxdz;

CovX = Jacobian * CovZ * Jacobian';
