%% B02DICPCovqzt Funcio que retorna la covariance de la estimacio de la pose
% d'un ICP (Iterative Closest Point)
% Parameters:
%   - c.m   -> Model Point Cloud (Ref)
%   - c.s   -> Readed Point Cloud (New)
%   - tx    -> X traslation estimated
%   - ty    -> Y traslation estimated
%   - tz    -> Z traslation estimated
%   - psi   -> Psi rotation estimated   (Rz)
%   - theta -> Theta rotation estimated (Ry)
%   - phi   -> Phi rotation estimated   (Rx)
%   - Covz  -> Points related covariance
% 
% Output:
%   - CovX   -> ICP3D Covariance Matrix [6x6]
% 
% Modified:
%   - Sergi Foix Salmerón
%   - PhD Student at IRI
    
%% code
function CovX = pt2pt_3D_Covariance_Polar( refPointCloud, newPointCloud, xyzRxRyRz, CovZ )
% Estimacio de la transformacio
tx      = xyzRxRyRz(1);
ty      = xyzRxRyRz(2);
tz      = xyzRxRyRz(3);
phi     = xyzRxRyRz(4);
theta   = xyzRxRyRz(5);
psi     = xyzRxRyRz(6);

Etot = 0;
Gtot = zeros(6,1);
EHess = zeros(6,6);

c.m = refPointCloud;
c.s = newPointCloud;

d2J_dxdei = zeros(6,size(c.s,2));
d2J_dxdej = zeros(6,size(c.s,2));

% These are the interval to approximate the derivatives with the increments.
eps_xyz = 0.00001;
eps_rot = deg2rad(0.01);

for ii=1:size(c.s,2)
    clc;
    fprintf('%d\n',ii);

    p_i = c.s(1:3,ii);
    p_j = c.m(1:3,ii);
    rho_i = norm(p_i);
    rho_j = norm(p_j);
    v_i = p_i / norm(p_i);
    v_j = p_j / norm(p_j);
    
    % The error function is the traditional point-to-point distance (squared)
    E_k = @(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, psi_) ...
            sum((...
                applyHRt2point( v_i*rho_i_, xyzRxRyRz2HRt([x_; y_; z_; phi_; theta_; psi_],'rad', 'pre')) ...
                - ...
                v_j*rho_j_ ...
                ).^2);
    
    %% First derivative
    gradEk = @(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, psi_) ...
        [ deriv(@(xx)E_k(rho_i_, rho_j_, xx, y_, z_, phi_, theta_, psi_), x_, eps_xyz); ...
          deriv(@(yy)E_k(rho_i_, rho_j_, x_, yy, z_, phi_, theta_, psi_), y_, eps_xyz); ...
          deriv(@(zz)E_k(rho_i_, rho_j_, x_, y_, zz, phi_, theta_, psi_), z_, eps_xyz); ...
          deriv(@(ph)E_k(rho_i_, rho_j_, x_, y_, z_, ph, theta_, psi_), phi_, eps_rot); ...
          deriv(@(tt)E_k(rho_i_, rho_j_, x_, y_, z_, phi_, tt, psi_), theta_, eps_rot); ...
          deriv(@(ps)E_k(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, ps), psi_, eps_rot)];

    %% Second derivative
    d2Ek_dx2 = @(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, psi_) ...
        [ deriv(@(xx)gradEk(rho_i_, rho_j_, xx, y_, z_, phi_, theta_, psi_), x_, eps_xyz) ...
          deriv(@(yy)gradEk(rho_i_, rho_j_, x_, yy, z_, phi_, theta_, psi_), y_, eps_xyz) ...
          deriv(@(zz)gradEk(rho_i_, rho_j_, x_, y_, zz, phi_, theta_, psi_), z_, eps_xyz) ...
          deriv(@(ph)gradEk(rho_i_, rho_j_, x_, y_, z_, ph, theta_, psi_), phi_, eps_rot) ...
          deriv(@(tt)gradEk(rho_i_, rho_j_, x_, y_, z_, phi_, tt, psi_), theta_, eps_rot) ...
          deriv(@(ps)gradEk(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, ps), psi_, eps_rot)];

    %% The noise has three components: epsilon_i, epsilon_j
    d2Ek_dxdei = @(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, psi_) ...
        deriv(@(ei)gradEk(ei, rho_j_, x_, y_, z_, phi_, theta_, psi_), rho_i_, eps_xyz);

    d2Ek_dxdej = @(rho_i_, rho_j_, x_, y_, z_, phi_, theta_, psi_) ...
        deriv(@(ej)gradEk(rho_i_, ej, x_, y_, z_, phi_, theta_, psi_), rho_j_, eps_xyz);
    
    Etot = Etot + E_k(rho_i, rho_j, tx, ty, tz, phi, theta, psi);
    Gtot = Gtot + gradEk(rho_i, rho_j, tx, ty, tz, phi, theta, psi);
    EHess = EHess + d2Ek_dx2(rho_i, rho_j, tx, ty, tz, phi, theta, psi);
    
    dei_ii = d2Ek_dxdei(rho_i, rho_j, tx, ty, tz, phi, theta, psi);
    dej_ii = d2Ek_dxdej(rho_i, rho_j, tx, ty, tz, phi, theta, psi);

    d2J_dxdei(:,ii) = d2J_dxdei(:,ii) + dei_ii;
    d2J_dxdej(:,ii) = d2J_dxdej(:,ii) + dej_ii;

end

E2J_dxdz = [ d2J_dxdei ];

Jacobian = inv(EHess)*E2J_dxdz;

CovX = Jacobian * CovZ * Jacobian';
