% MBICP3DCOVZTNUM 
% Compute numerically ICP covariance for the estimated pose  
%
%% Description:
%
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Scholarship
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 08/2009
%			Version 1.0
%
% Usage:
%       CovX = MbICP3DcovztNum(cloud_P,cloud_Pp, qMbICP, CovZ)
% Variables:
%
%	Inputs:
%       cloud_P  - 3xn matrix, first data set
%       cloud_Pp - 3xn matrix, second data set
%       qMbICP   - 6x1 vecctor,
%       CovZ     - (3*n)x(3*n) matrix,
% 
%   Output:
%       CovX     - ICP3D Covariance Matrix [6x6]
% 
% Notes:
% some similarities to the 
% variables are:
%
% cloud_P = dataP, z_t-1 c.m (Censi),S_ref p_2 (Biota),M (Nüchter),B (Aleksandr) 
%       , X (Besl)
% cloud_Pp =dataPp,z_t c.s (Censi),S_new p_1 (Biota),D (Nüchter), A (Aleksandr)
%       , P (Besl)

%% code
function CovX = MbICP3DcovztNum(cloud_P,cloud_Pp, qMbICP, CovZ,L)
% Estimacio de la transformacio
tx      = qMbICP(1);
ty      = qMbICP(2);
tz      = qMbICP(3);
tnx     = qMbICP(4);
tny     = qMbICP(5);
tnz     = qMbICP(6);

Etot = 0;
Gtot = zeros(6,1);
EHess = zeros(6,6);

c.m = cloud_P;
c.s = cloud_Pp;


d2J_dxdxz = zeros(6,size(c.s,2));
d2J_dxdyz = zeros(6,size(c.s,2));
d2J_dxdzz = zeros(6,size(c.s,2));

% These are the interval to approximate the derivatives with the increments.
eps_xyz = 0.00001;
eps_rot = 0.0001;

for ii=1:size(c.s,2)
    clc;
    fprintf('%d\n',ii);
    
    x_m = c.m(1,ii);
    y_m = c.m(2,ii);
    z_m = c.m(3,ii);
    x_z = c.s(1,ii);
    y_z = c.s(2,ii);
    z_z = c.s(3,ii);
    
    % The error function is the traditional point-to-point distance (squared)
    E_k = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_)(...
        MbICP3DFunction(tx,ty,tz,tnx,tny,tnz,x_m,y_m,z_m,x_z,y_z,z_z,L)...
        );
    
    %% First derivative
    gradEk = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_) ...
        [ deriv(@(xx)E_k(x_z_, y_z_, z_z_, xx, y_, z_, tnx_, tny_, tnz_), x_, eps_xyz); ...
          deriv(@(yy)E_k(x_z_, y_z_, z_z_, x_, yy, z_, tnx_, tny_, tnz_), y_, eps_xyz); ...
          deriv(@(zz)E_k(x_z_, y_z_, z_z_, x_, y_, zz, tnx_, tny_, tnz_), z_, eps_xyz); ...
          deriv(@(ph)E_k(x_z_, y_z_, z_z_, x_, y_, z_, ph, tny_, tnz_), tnx_, eps_rot); ...
          deriv(@(tt)E_k(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tt, tnz_), tny_, eps_rot); ...
          deriv(@(ps)E_k(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, ps), tnz_, eps_rot)];

    %% Second derivative
    d2Ek_dx2 = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_) ...
        [ deriv(@(xx)gradEk(x_z_, y_z_, z_z_, xx, y_, z_, tnx_, tny_, tnz_), x_, eps_xyz) ...
          deriv(@(yy)gradEk(x_z_, y_z_, z_z_, x_, yy, z_, tnx_, tny_, tnz_), y_, eps_xyz) ...
          deriv(@(zz)gradEk(x_z_, y_z_, z_z_, x_, y_, zz, tnx_, tny_, tnz_), z_, eps_xyz) ...
          deriv(@(ph)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, ph, tny_, tnz_), tnx_, eps_rot) ...
          deriv(@(tt)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tt, tnz_), tny_, eps_rot) ...
          deriv(@(ps)gradEk(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, ps), tnz_, eps_rot)];

    %% The noise has three components: etnzlon_i, etnzlon_j1, etnzlon_j2
    d2Ek_dxdexz = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_) ...
        deriv(@(exz)gradEk(exz, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_), x_z_, eps_xyz);

    d2Ek_dxdeyz = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_) ...
        deriv(@(eyz)gradEk(x_z_, eyz, z_z_, x_, y_, z_, tnx_, tny_, tnz_), y_z_, eps_xyz);
    
    d2Ek_dxdezz = @(x_z_, y_z_, z_z_, x_, y_, z_, tnx_, tny_, tnz_) ...
        deriv(@(ezz)gradEk(x_z_, y_z_, ezz, x_, y_, z_, tnx_, tny_, tnz_), z_z_, eps_xyz);

    Etot = Etot + E_k(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);
    Gtot = Gtot + gradEk(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);
    EHess = EHess + d2Ek_dx2(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);
    
    dxz_ii = d2Ek_dxdexz(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);
    dyz_ii = d2Ek_dxdeyz(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);
    dzz_ii = d2Ek_dxdezz(x_z, y_z, z_z, tx, ty, tz, tnx, tny, tnz);

    d2J_dxdxz(:,ii) = d2J_dxdxz(:,ii) + dxz_ii;
    d2J_dxdyz(:,ii) = d2J_dxdyz(:,ii) + dyz_ii;
    d2J_dxdzz(:,ii) = d2J_dxdzz(:,ii) + dzz_ii;

end

E2J_dxdz = [ d2J_dxdxz  d2J_dxdyz d2J_dxdzz ];

Jacobian = inv(EHess)*E2J_dxdz;

CovX = Jacobian * CovZ * Jacobian';
