%% USEANN
% Use the ANN library to find the correspondeces for the point-to-point and
% point-toplane metrics. Returning the correspondent querypoint and its
% nearest neighbor
%
%% Description
%
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Fellow
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 02/2009
%
% This code has been developed as part of the EU URUS Project
% FP6-IST-045062
%           
%% Usage
%   nnidx =useANNNormals2(queryPts,refPts[,OPTIONS])
% Variables
%   Inputs:
%       queryPts 	- a 3 x n matrix, points searching its neighbors
%       refPts    - a 3 x n matrix, points in which neigbors are searched
%   Options;
%       'refPlaneError'  - Maximum distance for filtering ANN
%       'NNN'            - Number of nearest neighbors 
%
%   errometric      - a struct
%       errormetric.type - a boolean, 0 pt2pt, 1 hybrid
%       errometric.maxD2Planes - depends on the unit you are working.
%       errometric.errTrPlanes - depends on the unit you are working.
%   filter          - a struct
%       filter.normals.do   - a boolena, use normal's filtering
%       filter.normals.limtheta - a real, Max angle between normals 
%                                 in radians
%   
%
% Outputs:
%       ansPoints - a n x 7 matrix, paired points and it's distance
%                   - row 1 to 3, same as query points
%                   - row 4 to 6, NN 
%                   - row 7, 
%       nnindx    -  Output the matrix containg the index for the matching
%                  Column 1 : query/model indexs
%                  Column 2 : reference indexs
%

%% CODE
function  nnpts = useANN(queryPts,refPts,errorMetric,filters,...
    varargin)



% errometric.type = 0;         % 0 pt2pt, 1 hybrid
% errometric.maxD2Planes = 3;  % Maximum distance constriain to planes
% errometric.errTrPlanes = 10000; % Adjust error 4 planes


[r1 c1] = size(queryPts);
r2  = size(refPts,1);

if r1~=r2
    error('Invalid dimension compatibility between reference and query points')
end

if (r1 < 3 || r2 < 3)
   error('Mismatching dimentions for the query or reference data')
end

narg = length(varargin);

if (mod(narg, 2))
  error('Each option must be a string/value pair.');
end

%% Default values
NNN = 1;

refPlaneError = [];
for i=1:2:narg
    switch char(varargin(i))
        case 'refPlaneError'
            refPlaneError = varargin{i+1};
    end
end

% if isempty(refPlaneError) && strcmp(errorMetric,'1')
%     error('Whe using planes, you ought to pass a matrix containing the plane adjust error')
% end

% ANN return the index (nnidx) for the queryPts correspondet with the
[annidx dANN]= annquery(refPts.points, queryPts.points, NNN);

% Notas: 
% A  -cuando los datos son casi = o = ANN no puede regresar la distancia
% B - verificar que exista nnidx si no existe mostrar un mensaje que
% diga que no se pudieron econtrar vecindades entre las dos nubes
% de puntos
% fprintf('Using ANN 0, points \n');

nnidx = zeros(c1,3) ;

for i =1:c1
    % recovering the querypoints 
    nnidx(i,1) = i;    
end
nnidx(:,2) = annidx';

% Euclidian ditance between the two i points
nnidx(:,3) = dANN';      
% Filter the repeated points retuned by ANN only for the pt2pt metric
nnidx =sortrows(nnidx,2);
if ~errorMetric.type 
    nnidx = filIndexn(nnidx(:,1),nnidx(:,2),nnidx(:,3),maxDist); 
end


% Normals filter
if isfield(filters,'normals')
    j = 1;
    if (isfield(refPts,'normals')&& isfield(queryPts,'normals') )
      for i = 1:size(nnidx,1)
          theta = unitNormalsAngle(queryPts.normals(:,nnidx(i,1)),...
              refPts.normals(:,nnidx(i,2)));
          if theta > pi/2
              theta = theta-pi;
          end
          if theta <= str2double(filters.normals)
              nnidxNew(j,:) = nnidx(i,1:3); 
              % newData.normal(j,:) = tData(1).normal(i,:);
              j = j+1;
          end
      end
    else
      error('Missing field normals in structs scan3Dxxx')
    end
else
  nnidxNew =  nnidx;
end


if strcmp(errorMetric.type, '0')  
    % USE POINTS pt2pt metric
    % fprintf('Using pt2pt metric \n')
    nnpts(:,1:3) = queryPts.points(:,nnidxNew(:,1))';
    nnpts(:,4:6) = refPts.points(:,nnidxNew(:,2))';
    nnpts(:,7) = nnidxNew(:,3);
else
  if ~(isfield(errorMetric,'errorAdjPlanes'))
      error('Missing parameter for error metric when using planes/hybrid')
  end
  % USE PLANES, use pt2pt and pt2plane hybrid metric 
  % or pt2plane only if config.errometric.errTrPlanes is too big
  % fprintf('Using hybrid metric \n')
  j = 1;
  nnpts(1,:) = zeros(1,7);
  for i=1:2:narg
      switch char(varargin(i))
          case 'etr'
              errorMetric.errorAdjPlanes = varargin{i+1};
      end
  end
  r3 = size(nnidxNew,1);
  maxD2Planes = str2double(filters.maxD2Planes);
  errorAdjPlanes = str2double(errorMetric.errorAdjPlanes);
  for i = 1:r3
    % Filtrar, las distancias maximas de los puntos  los planos
    if isfield(filters,'maxD2Planes')
      if nnidxNew(i,3) < maxD2Planes
          % Chose the correpondece between pt2pt or pt2plane
          if refPlaneError(nnidxNew(i,2)) > errorAdjPlanes
              % pt2pt metric
              % Keep the closes point in the reference as the
              % correspondent point to the query point
              ansPt = refPts.points(:,nnidxNew(i,2))';
              d =  nnidxNew(i,3);
              % m = 0
          else
              % pt2plane metric
              % Find the closes point in the plane to the query point
              n = scan3Dref.normals(nnidxNew(i,2),:);
              p = refPts.points(:,nnidxNew(i,2))';
              [ansPt d]= pointPlane2Point(n',p',queryPts.points(:,nnidxNew(i,1)));
              ansPt = ansPt';
          end
          nnpts(j,:) = [queryPts.points(:,nnidxNew(i,1))',ansPt, abs(d)];
          j = j+1;
      end
    else
      if refPlaneError(nnidxNew(i,2)) > errorAdjPlanes
        % pt2pt metric
        % Keep the closes point in the reference as the
        % correspondent point to the query point
        ansPt = refPts.points(:,nnidxNew(i,2))';
        d =  nnidxNew(i,3);
        % m = 0
      else
        % pt2plane metric
        % Find the closes point in the plane to the query point
        n = refPts.normals(nnidxNew(i,2),:);
        p = refPts.points(:,nnidxNew(i,2))';
        [ansPt d]= pointPlane2Point(n',p',queryPts.points(:,nnidxNew(i,1)));
        ansPt = ansPt';
      end
      nnpts(j,:) = [queryPts.points(:,nnidxNew(i,1))',ansPt, abs(d)];
      j = j+1;
    end
  end
end

if size(nnpts,1)<=100
  error('No enough correspondeces')
end
        
end
