% ICP3DRUN
% Metric-based ICP hybrid algorithm, using NN library for matching in
% Euclidean space
%
%% Description
%
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Fellow
%           Juan Andrade Cetto
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 08/2009
%           
%
% This code has been developed as part of the EU URUS Project
% FP6-IST-045062
% 
% Usage
%   [newRT Err]= icp3Drun(refPts, queryPts,pose[,OPTIONS])
%
% Varibles:
%   Input:
%     refPts    - 3xn matrix, reference points, old cloud
%     queryPts  - 3xm matrix, query points, or new cloud
%     d_tat1 - 6x1 vector, displacement from t to t_1(queryPts) the point clouds
%
%   Options:
%       'Plot'        - integere, plotnumber 
%       'PlotBGC'     - valid color valu, plot back graund color
%       'ColseWhenDone' -
%       'NNSearch'    - a string 'GLTree',      ...
%       'MatchMetric' - bool  0-pts(DEFAULT), 1-pts and planes, or only
%                       planes depends on 'Planes' fielda values
%       'Normals'     - a cell,{                               ...
%                       'refNormals',                            ...
%                        3xn matrix- reference Scan 3 DNormals'), ...
%                       'queryNormals',                          ...
%                        3xm matrix- query Scan 3 DNormals'),    ...
%                       'maxAngle', a +real deg2rad(30)'DEAFULT'}
%       'Planes'        - a cell{                                      ...
%                       'refPlanesError',                              ...
%                        1xn vector-reference scan planes error'),...
%                       'planesTreshold', a +real}
%       'MaxIterations' - a +integer, 25 DEFAULT
%       'MinIterations' - a +integer, 3 DEFAULT
%       'MSETreshold'   - a +integer,10 DEFAULT
%       'MSEType'       - a string, 'Mb'(DEFAULT), 'SVD', 'Quat'
%       'Mb_L'          - a +integer  10);
%
%   Output:
%       newRT       - a 4 x 4 matrix, computed homogeneus transformation matrix
%       Err         - structure, returning the computed errors
%                       Err.dk2  - a real, median square error
%                       Err.covM - a 6 x6 matrix, covariance matrix
%                       Err.muV  - a 6 x 1 vector, median error vector
%
%   Global
%       config.conv.maxiter - an integer, maximun number of expected
%                             iterations to finish the algorithm


%% Code

function [newRT covM varargout]= icp3Drun(refPts, queryPts,d_tat1,...
                                            varargin)




tic;
% DEFAULTS

% *************************************************************************
chkvarargin(varargin,2);
[ plot,        ...
  bgc,         ...
  closewhendone, ...
  nnsearch,    ...
  matchmetric, ...
  normalsparam, ...
  planesparam, ...
  maxiter,     ...
  miniter,     ...
  mindk2,      ...
  mseTreshold, ...
  mseType,     ...
  L,           ...
  textdisp,    ...
  dsigma,      ...
  dfixed,      ...
  covType,     ...
  mincorr,     ...
  boxsize,     ...
  savematch]  = process_options(varargin,...
                    'Plot',0,                 ...
                    'PlotBGC', 'w',           ...
                    'ColseWhenDone',0,        ...
                    'NNSearch','GLTree',      ...
                    'MatchMetric', 0,         ...
                    'Normals', [],      ...
                    'Planes', [],             ...
                    'MaxIterations',25,       ...
                    'MinIterations', 3,       ...
                    'MinMSE', 0.1,            ...
                    'MSETreshold', 100,       ...
                    'MSEType', 'Mb',          ...
                    'Mb_L' ,  30,             ...
                    'TextDisplay',0,          ...
                    'DistanceFilterSigma',[], ...
                    'DistanceFilterFixed',[], ...
                    'Covariance', 'MSE',    ...
                    'MinCorrespondences',[],  ...
                    'BoxSize', 1,...
                    'SaveMatch',{'SaveDir', [], 'ID', []});

      % Patch 201505 , por el error Undefined function or variable 'savedirmatch'.
      savedirmatch = [];
                  
if ~isempty(normalsparam)
  vars{1} = 'RefNormals';
  vars{2} = 'QueryNormals' ;
  vars{3} = 'MaxAngle';
  x = isstrincell(vars,normalsparam);
  if x(1)
    refNormals = normalsparam{x(1)*2};
    if ~isempty(refNormals)
      if size(refPts,2)~=size(refNormals,2)
        error('Missmatching dimentions between reference points and normals')
      end
      if x
        if x(2)
          queryNormals = normalsparam{x(2)*2};
          if isempty(queryNormals)
            error('Empty queryNormals unable to do normal filter')
          end
          if size(queryPts,2)~=size(queryNormals,2)
            error('Missmatching dimentions between query points and query normals')
          end
          if ~isstruct(queryNormals)
            queryNormals = rotateData3D(queryNormals,rxryrz2R(d_tat1(4:6)));
          else
            queryNormals.h = rotateData3D(queryNormals.h,rxryrz2R(d_tat1(4:6)));
            queryNormals.v = rotateData3D(queryNormals.v,rxryrz2R(d_tat1(4:6)));
          end
        else
          if isstruct(refPts)
            queryNormals.h = [];
            queryNormals.v = [];
          else
            queryNormals = [];
          end
        end
        if x(3)
          maxAngle = normalsparam{x(3)*2};
        else
          maxAngle = [];
        end
      else
        if x(3)
          error('Missing reference normals, for doing')
        end
        if isstruct(refPts)
          queryNormals.h =[];
          maxAngle =[];
          queryNormals.v =[];
        else
          queryNormals =[];
          maxAngle =[];
        end
      end
    end
  else
    msg = ['Invalid parameter for Normals, valid parameters are: \n' ...
            'refNormals, queryNormals or maxAngle'];
    error(msg)
  end
else
  if isstruct(refPts)
    refNormals.h = [];
    queryNormals.h =[];
    refNormals.v = [];
    queryNormals.v =[];
    maxAngle = [];
  else
    refNormals = [];
    queryNormals =[];
    maxAngle =[];
  end
end 

if ~isempty(planesparam) && ~isempty(refNormals)
  vars{1} = 'RefPlanesError';
  vars{2} = 'PlanesTreshold';
  x = isstrincell(vars,planesparam);
  if x(1)
    refPlanesError = planesparam{x(1)*2};
    if x(2)
      planesTreshold = planesparam{x(2)*2};
    else
       planesTreshold = 100000;
    end
  end
else isempty(planesparam) 

  if isstruct(refPts)
    refPlanesError.h =[];
    refPlanesError.v =[];
  else
    refPlanesError = [];
  end
  planesTreshold =[];
end
clear varargin
% Validating data size

Rt(1:maxiter) = struct('H',[]);

% *************************************************************************
% ***                            ICP                                    ***
% *************************************************************************
%  Initialazing homogenues matrix Rt
Rt(1).H = xyzrxryrz2HRt(d_tat1);
his.e(1).covM = zeros(6); %Original covariance
Err.dk2 = [];
err(1,:) = [0  0];
H_ = eye(4);
iter = 1; loop = 0; lstdk2 = 100000;

% *************************************************************************
% ***                           Previous                                ***
% *************************************************************************
%tmp_ = queryPts;
if isstruct(queryPts)
  fprintf('Matching Semantic:\n ');
  queryPts.h = applyHRt2data(queryPts.h,xyzrxryrz2HRt(d_tat1));
  queryPts.v = applyHRt2data(queryPts.v,xyzrxryrz2HRt(d_tat1));
  rInfo.v =[];
  rIndex.v =[];
  rInfo.h =[];
  rIndex.h =[];
else
  fprintf('Matching:\n ')
  queryPts = applyHRt2data(queryPts,xyzrxryrz2HRt(d_tat1));
  rInfo =[];
  rIndex =[];
end

if matchmetric == 0
        fprintf('\tMetric->points\n');
elseif matchmetric == 1
        fprintf('\tMetric->points and planes,\n ');
end

switch mseType 
  case {'Mb','SVD','Quat'}
    fprintf('\tMSE method->%s\n ',mseType);
  otherwise
   error('Invalid MSE method') 
end

if strcmp(nnsearch,'BNN')
  delta = 0.1;
  % Build box structure for the reference point it is only need once, as
  % long as box size is the same
  if isstruct(refPts)
    refBox.h = getBoxStructMine(refPts.h,boxsize,delta);
    refBox.v =getBoxStructMine(refPts.v,boxsize,delta);
  else
    
    if ~isempty(savematch{2})
      vars{1} = 'SaveDir';
      vars{2} =  'ID';
      x = isstrincell(vars,savematch);
      %x = isstrincell(vars,normalsparam); %2015-05
      if  x(1)  
        if (2)
          refBox = getBoxStructMine(refPts,boxsize,delta,...
                    savematch{x(1)*2},savematch{x(2)*2});
        end
      end
    else
      refBox = getBoxStructMine(refPts,boxsize,delta);
    end
  end
else
  if isstruct(refPts)
    refBox.h =[];refBox.v =[];
  else
    refBox =[];
  end
end

fprintf('\tNN search->%s\n',nnsearch);

if plot
  % Plots the Reference(blue) and Model(red) clouds
   icp3DPlotsA(refPts,queryPts,plot);
end 

% *************************************************************************
% ***                         Algorithm                                 ***
% *************************************************************************


if ~textdisp
  fprintf('Iterations %s:',num2str(0, '%04u'))
end

% fr = 0.1;
while (loop == 0) % < 1
    if textdisp
      fprintf('Iteration : %3.0f ',iter);
    else
      fprintf('\b\b\b\b\b')
      fprintf('%s:',num2str(iter, '%04u'))
    end

    if mod(iter,2)
      fIter= maxiter;
      if isstruct(refPts)
        % Semantic icp, horizontal and vertical
        [cXrXqs.h flags.h cInx.h]= icp3DMatching( refPts.h,     ...
                                  queryPts.h,                   ...
                                  'RefNormals',refNormals.h,    ...
                                  'QueryNormals',queryNormals.h , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError.h,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter, ...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L, ...
                                  'ReferenceBox',refBox.h);
        % implementar el cambio de origen de caja ox,oy,oz,0
        [cXrXqs.v flags.v cInx.v]= icp3DMatching( refPts.v,     ...
                                  queryPts.v,                   ...
                                  'RefNormals',refNormals.v,    ...
                                  'QueryNormals',queryNormals.v , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError.v,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter, ...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L,                       ...
                                  'ReferenceBox',refBox.v);
        flagC = flags.h+flags.v;
        cXrXq = [cXrXqs.h cXrXqs.v];
      else
        [cXrXq flagC]= icp3DMatching( refPts,...
                                  queryPts,...
                                  'RefNormals',refNormals,      ...
                                  'QueryNormals',queryNormals , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter,...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L,                       ...
                                  'ReferenceBox',refBox, ...
                                  'Savematch',savematch, ...
                                  'Iteration',iter);
      end
    end
    % Todo verify y if there are enough correpondeces
    if flagC
      % correspondences from reference points
      cXr =cXrXq(1:3,:);
      % correspondences from query points
      cXq =cXrXq(4:6,:);
      % Find the ek mean error for convergence analysis:
      err(iter+1,1) = error2(cXq ,cXr,'Type', mseType,'L',L);
      % pregunta por que da casi cero siempre
      % Minimization Step : 
      H = icp3DMinimization(cXr, cXq,'MSEType',mseType,'L',L);
      Rt(iter+1).H = H*Rt(iter).H;
      H_ = H_*H;
      % Updating step
      if isstruct(queryPts)
        queryPts.h = applyHRt2data(queryPts.h,H);
        queryPts.v = applyHRt2data(queryPts.v,H);
      else
        queryPts = applyHRt2data(queryPts,H);
      end
      cXq = applyHRt2data(cXq,H);
      if ~isempty(queryNormals) 
        if isstruct(queryNormals)
          if ~isempty(queryNormals.h)
            queryNormals.h = rotateData3D(queryNormals.h,H(1:3,1:3));
            queryNormals.v = rotateData3D(queryNormals.v,H(1:3,1:3));
          end
        else
          queryNormals = rotateData3D(queryNormals,H(1:3,1:3));
        end
        if maxAngle > 0.2618 % dinamic angle adjustment
           maxAngle =maxAngle - 1.05*max(abs(R2rxryrz(H(1:3,1:3))));
        end
      end
  
      % Find the dk mean square error [Besl92] convergence theorem:
      err(iter+1,2) = error2(cXq ,cXr,'Type', mseType,'L',L);
      if textdisp
          fprintf(' ek2 = %2.4f, dk2 = %2.4f ,',err(iter+1,1), err(iter+1,2))
          fprintf(' ek2-dk2 = %2.4f \n',err(iter+1,1)-err(iter+1,2) )
      end
      % Convergence Theorem
     [loop lstdk2]= convTheo( err(iter+1,1),...
                              err(iter+1,2),...
                              lstdk2,mindk2,...
                              mseTreshold,  ...
                              iter, maxiter);
      iter = iter + 1;

    else
      loop =6;
    end
end


fprintf('\t %d ', iter-1)
% *************************************************************************



% Execution time
fprintf('DONE\n ');
fprintf(' ek2 = %2.6f, dk2 = %2.6f ,',err(iter,1), err(iter,2))
          fprintf(' ek2-dk2 = %2.6f \n',err(iter,1)-err(iter,2) )
showElapsedTime(toc);
newPose.H = H_*xyzrxryrz2HRt(d_tat1);

% 


% The new pose is returned according with ICP's updates
newRT = newPose.H;
% newRT = Rt(iter).H;
switch covType
  case 'Naive'
    covM = naiveCovariance(cXr ,cXq);
  case 'MSE'
    covM = computeCov(cXr,cXq,H); 
end
if plot
  icp3DPlotsC(queryPts,plot,cXr,cXq,...
              newPose.H,covM)
end
% Means Square Error
Err.dk2 = err(iter,2);
% Covariance from scan matching
% TODO: to compute the covariance according with the minimization used
%Err.covM = naiveCovariance(cXr,cXq);
Err.loop = loop;


%%

if textdisp
    fprintf('Convergence %d :\n ',loop)
end
if nargout == 3 
    varargout(1) = {err};
end

if ~textdisp
    fprintf('\n')
end
