% ICP3DRUN
% Metric-based ICP hybrid algorithm, using NN library for matching in
% Euclidean space
%
%% Description
%
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Fellow
%           Juan Andrade Cetto
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 08/2009
%           
%
% This code has been developed as part of the EU URUS Project
% FP6-IST-045062
% 
% Usage
%   [newRT Err]= icp3Drun(refPts, queryPts,pose[,OPTIONS])
%
% Varibles:
%   Input:
%     refPts    - 3xn matrix, reference points, old cloud
%     queryPts  - 3xm matrix, query points, or new cloud
%     d_tat1 - 6x1 vector, displacement from t to t_1(queryPts) the point clouds
%
%   Options:
%       'Plot'        - integere, plotnumber 
%       'PlotBGC'     - valid color valu, plot back graund color
%       'ColseWhenDone' -
%       'NNSearch'    - a string 'GLTree',      ...
%       'MatchMetric' - bool  0-pts(DEFAULT), 1-pts and planes, or only
%                       planes depends on 'Planes' fielda values
%       'Normals'     - a cell,{                               ...
%                       'refNormals',                            ...
%                        3xn matrix- reference Scan 3 DNormals'), ...
%                       'queryNormals',                          ...
%                        3xm matrix- query Scan 3 DNormals'),    ...
%                       'maxAngle', a +real deg2rad(30)'DEAFULT'}
%       'Planes'        - a cell{                                      ...
%                       'refPlanesError',                              ...
%                        1xn vector-reference scan planes error'),...
%                       'planesTreshold', a +real}
%       'MaxIterations' - a +integer, 25 DEFAULT
%       'MinIterations' - a +integer, 3 DEFAULT
%       'MSETreshold'   - a +integer,10 DEFAULT
%       'MSEType'       - a string, 'Mb'(DEFAULT), 'SVD', 'Quat'
%       'Mb_L'          - a +integer  10);
%
%   Output:
%       newRT       - a 4 x 4 matrix, computed homogeneus transformation matrix
%       Err         - structure, returning the computed errors
%                       Err.dk2  - a real, median square error
%                       Err.covM - a 6 x6 matrix, covariance matrix
%                       Err.muV  - a 6 x 1 vector, median error vector
%
%   Global
%       config.conv.maxiter - an integer, maximun number of expected
%                             iterations to finish the algorithm


%% Code

function [newRT covM varargout]= icp3Drun(refPts, queryPts,d_tat1,...
                                            varargin)




tic;
% DEFAULTS

% *************************************************************************
chkvarargin(varargin,2);
[ plot        ...
  bgc         ...
  closewhendone ...
  nnsearch    ...
  matchmetric ...
  normalsparam...
  planesparam ...
  maxiter     ...
  miniter     ...
  mindk2      ...
  mseTreshold ...
  mseType     ...
  L           ...
  textdisp    ...
  dsigma      ...
  dfixed      ...
  covType     ...
  mincorr     ...
  boxsize]  = process_options(varargin,...
                    'Plot',0,                 ...
                    'PlotBGC', 'w',           ...
                    'ColseWhenDone',0,        ...
                    'NNSearch','GLTree',      ...
                    'MatchMetric', 0,         ...
                    'Normals', [],      ...
                    'Planes', [],             ...
                    'MaxIterations',25,       ...
                    'MinIterations', 3,       ...
                    'MinMSE', 0.1,            ...
                    'MSETreshold', 100,       ...
                    'MSEType', 'Mb',          ...
                    'Mb_L' ,  30,             ...
                    'TextDisplay',0,          ...
                    'DistanceFilterSigma',[], ...
                    'DistanceFilterFixed',[], ...
                    'Covariance', 'Naive',    ...
                    'MinCorrespondences',[],  ...
                    'BoxSize', 2);
                
if ~isempty(normalsparam)
  vars{1} = 'RefNormals';
  vars{2} = 'QueryNormals' ;
  vars{3} = 'MaxAngle';
  x = isstrincell(vars,normalsparam);
  if x(1)
    refNormals = normalsparam{x(1)*2};
    if ~isempty(refNormals)
      if size(refPts,2)~=size(refNormals,2)
        error('Missmatching dimentions between reference points and normals')
      end
      if x
        if x(2)
          queryNormals = normalsparam{x(2)*2};
          if isempty(queryNormals)
            error('Empty queryNormals unable to do normal filter')
          end
          if size(queryPts,2)~=size(queryNormals,2)
            error('Missmatching dimentions between query points and query normals')
          end
          if ~isstruct(queryNormals)
            queryNormals = rotateData3D(queryNormals,rxryrz2R(d_tat1(4:6)));
          else
            queryNormals.h = rotateData3D(queryNormals.h,rxryrz2R(d_tat1(4:6)));
            queryNormals.v = rotateData3D(queryNormals.v,rxryrz2R(d_tat1(4:6)));
          end
        else
          if isstruct(refPts)
            queryNormals.h = [];
            queryNormals.v = [];
          else
            queryNormals = [];
          end
        end
        if x(3)
          maxAngle = normalsparam{x(3)*2};
        else
          maxAngle = [];
        end
      else
        if x(3)
          error('Missing reference normals, for doing')
        end
        if isstruct(refPts)
          queryNormals.h =[];
          maxAngle =[];
          queryNormals.v =[];
        else
          queryNormals =[];
          maxAngle =[];
        end
      end
    end
  else
    msg = ['Invalid parameter for Normals, valid parameters are: \n' ...
            'refNormals, queryNormals or maxAngle'];
    error(msg)
  end
else
  if isstruct(refPts)
    refNormals.h = [];
    queryNormals.h =[];
    refNormals.v = [];
    queryNormals.v =[];
    maxAngle = [];
  else
    refNormals = [];
    queryNormals =[];
    maxAngle =[];
  end
end 

if ~isempty(planesparam) && ~isempty(refNormals)
  vars{1} = 'RefPlanesError';
  vars{2} = 'PlanesTreshold';
  x = isstrincell(vars,planesparam);
  if x(1)
    refPlanesError = planesparam{x(1)*2};
    if x(2)
      planesTreshold = planesparam{x(2)*2};
    else
       planesTreshold = 100000;
    end
  end
else isempty(planesparam) 
  refPlanesError = [];
  planesTreshold =[];
end
clear varargin
% Validating data size

Rt(1:maxiter) = struct('H',[]);

% *************************************************************************
% ***                            ICP                                    ***
% *************************************************************************
%  Initialazing homogenues matrix Rt
Rt(1).H = xyzrxryrz2HRt(d_tat1);
his.e(1).covM = zeros(6); %Original covariance
Err.dk2 = [];
err(1,:) = [0  0];
H_ = eye(4);
iter = 1; loop = 0; lstdk2 = 100000;

% *************************************************************************
% ***                           Previous                                ***
% *************************************************************************
%tmp_ = queryPts;
if isstruct(queryPts)
  fprintf('Matching Semantic: ');
  queryPts.h = applyHRt2data(queryPts.h,xyzrxryrz2HRt(d_tat1));
  queryPts.v = applyHRt2data(queryPts.v,xyzrxryrz2HRt(d_tat1));
  rInfo.v =[];
  rIndex.v =[];
  rInfo.h =[];
  rIndex.h =[];
else
  fprintf('Matching: ')
  queryPts = applyHRt2data(queryPts,xyzrxryrz2HRt(d_tat1));
  rInfo =[];
  rIndex =[];
end

if matchmetric == 0
        fprintf('metric->points,');
elseif matchmetric == 1
        fprintf('metric->points and planes, ');
end

switch mseType 
  case {'Mb','SVD','Quat'}
    fprintf('MSE method->%s, ',mseType);
  otherwise
   error('Invalid MSE method') 
end

fprintf('NN search->%s\n',nnsearch);

if plot
  % Plots the Reference(blue) and Model(red) clouds
  if isstruct(refPts)
    plotData3D(refPts.h,'FigureNumber',plot, ...
              'bgc',bgc,  ...
              'Subfigure',[1 2 1]);
    plotData3D(queryPts.h,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 1],      ...
              'Color','r');
    plotData3D(refPts.h,'FigureNumber',plot, ...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2]);
    plotData3D(refPts.v,'FigureNumber',plot, ...
            'bgc',bgc,  ...
            'Subfigure',[1 2 1]);
    plotData3D(queryPts.v,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 1],      ...
              'Color','r');
    plotData3D(refPts.v,'FigureNumber',plot, ...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2]); 
  else
    plotData3D(refPts,'FigureNumber',plot, ...
            'bgc',bgc,  ...
            'Subfigure',[1 2 1]);
    plotData3D(queryPts,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 1],      ...
              'Color','r');
    plotData3D(refPts,'FigureNumber',plot, ...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2]); 
  end
end 

% *************************************************************************
% ***                         Algorithm                                 ***
% *************************************************************************


if ~textdisp
  fprintf('Iterations %s:',num2str(0, '%04u'))
end

fr = 0.1;
while (loop == 0) % < 1
    if textdisp
      fprintf('Iteration : %3.0f ',iter);
    else
      fprintf('\b\b\b\b\b')
      fprintf('%s:',num2str(iter, '%04u'))
    end
    % Find the correspondences 
    mIter = maxiter;

    if iter <= mIter
      fIter= maxiter;
%       if dsigma>3.5
%         dsigma = dsigma-fr*dsigma;
%       end
      if isstruct(refPts)
        [cXrXqs.h flags.h cInx.h]= icp3DMatching2( refPts.h,     ...
                                  queryPts.h,                   ...
                                  'RefNormals',refNormals.h,    ...
                                  'QueryNormals',queryNormals.h , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError.h,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter, ...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L);
        % implementar el cambio de origen de caja ox,oy,oz,0
        [cXrXqs.v flags.v cInx.v]= icp3DMatching2( refPts.v,     ...
                                  queryPts.v,                   ...
                                  'RefNormals',refNormals.v,    ...
                                  'QueryNormals',queryNormals.v , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError.v,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter, ...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L);
        flagC = flags.h+flags.v;
        cXrXq = [cXrXqs.h cXrXqs.v];
      else
        [cXrXq flagC]= icp3DMatching2( refPts,...
                                  queryPts,...
                                  'RefNormals',refNormals,      ...
                                  'QueryNormals',queryNormals , ...
                                  'MaxAngle',maxAngle,          ...
                                  'PlanesError', refPlanesError,...
                                  'MaxPlanesErr', planesTreshold, ...
                                  'NNSearch',nnsearch,          ...
                                  'MatchMetric', matchmetric,   ...
                                  'DistanceFilterSigma',dsigma, ...
                                  'DistanceFilterFixed',dfixed, ...
                                  'MinCorrespondences',mincorr, ...
                                  'UseFilterBeforeIteration',fIter,...
                                  'CurrentIteration',iter,      ...
                                  'BoxSize',boxsize,            ...
                                  'DistanceMetric', mseType,    ...
                                  'L', L);
      end
    end
    % Todo verify y if there are enough correpondeces
    if flagC
      % correspondences from reference points
      cXr =cXrXq(1:3,:);
      % correspondences from query points
      cXq =cXrXq(4:6,:);
      % Find the ek mean error for convergence analysis:
      err(iter+1,1) = error2(cXq ,cXr,'Type', mseType,'L',L);
      % pregunta por que da casi cero siempre
      % Minimization Step : 
      H = icp3DMinimization(cXr, cXq,'MSEType',mseType,'L',L);
      Rt(iter+1).H = H*Rt(iter).H;
      H_ = H_*H;
      % Updating step
      if isstruct(queryPts)
        queryPts.h = applyHRt2data(queryPts.h,H);
        queryPts.v = applyHRt2data(queryPts.v,H);
      else
        queryPts = applyHRt2data(queryPts,H);
      end
      cXq = applyHRt2data(cXq,H);
      if ~isempty(queryNormals) 
        if isstruct(queryNormals)
          if ~isempty(queryNormals.h)
            queryNormals.h = rotateData3D(queryNormals.h,H(1:3,1:3));
            queryNormals.v = rotateData3D(queryNormals.v,H(1:3,1:3));
          end
        else
          queryNormals = rotateData3D(queryNormals,H(1:3,1:3));
        end
        if maxAngle > 0.2618 % dinamic angle adjustment
           maxAngle =maxAngle - 1.05*max(abs(R2rxryrz(H(1:3,1:3))));
        end
      end
  
      % Find the dk mean square error [Besl92] convergence theorem:
      err(iter+1,2) = error2(cXq ,cXr,'Type', mseType,'L',L);
      % Compute the covariance . TODO, compute covariance only when icp nas
      % converged to safe computational time
      switch covType
        case 'Naive'
          his.e(iter+1).covM = naiveCovariance(cXr ,cXq);
        case 'MSE'
          % TODO
         his.e(iter+1).covM = computeCov(cXr,cXq,ccXr,ccXq,mseType,'L',L); 
      end
      % TODO, actualizar el codigo para usar errorCov2 y crear erroMu para el
      % Historial de muV
       if textdisp
          fprintf(' ek2 = %2.4f, dk2 = %2.4f ,',err(iter+1,1), err(iter+1,2))
          fprintf(' ek2-dk2 = %2.4f \n',err(iter+1,1)-err(iter+1,2) )
      end
      % Convergence Theorem
     [loop lstdk2]= convTheo( err(iter+1,1),...
                              err(iter+1,2),...
                              lstdk2,mindk2,...
                              mseTreshold,  ...
                              iter, maxiter);
      iter = iter + 1;
     % TODO: si diverge desde un pricipio regresar Rt(1)
%     if plot
%   if iter >74
%        if isstruct(queryPts)
%           plotData3D(queryPts.h,'FigureNumber',plot,...
%                     'bgc',bgc,  ...
%                     'Subfigure',[1 2 2],      ...
%                     'Color', 'r');
%           plotData3D(queryPts.v,'FigureNumber',plot,...
%                     'bgc',bgc,  ...
%                     'Subfigure',[1 2 2],      ...
%                     'Color', 'r');
%         else
%          plotData3D(queryPts,'FigureNumber',plot,...
%                     'bgc',bgc,  ...
%                     'Subfigure',[1 2 2],      ...
%                     'Color', 'r');
%         end
%        plotCorrespondences3D([cXr;cXq],...
%                   'Subfigure',[1 2 2]);
%        plotcov3(Rt(iter).H(1:3,4),his.e(iter).covM(1:3,1:3));
%        hold off;
%        cameratoolbar;
%     end
    else
      loop =6;
    end
end


fprintf('\t %d ', iter-1)
% *************************************************************************



% Execution time
fprintf('DONE\n ');
fprintf(' ek2 = %2.6f, dk2 = %2.6f ,',err(iter,1), err(iter,2))
          fprintf(' ek2-dk2 = %2.6f \n',err(iter,1)-err(iter,2) )
showElapsedTime(toc);
newPose.H = H_*xyzrxryrz2HRt(d_tat1);


if plot
%   plotData3D(applyHRt2data(tmp_,newPose.H),...
%             'FigureNumber',plot,...
%             'bgc',bgc,  ...
%             'Subfigure',[1 2 2],      ...
%             'Color', 'g');
  if isstruct(queryPts)
    plotData3D(queryPts.h,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2],      ...
              'Color', 'r');
    plotData3D(queryPts.v,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2],      ...
              'Color', 'r');
  else
   plotData3D(queryPts,'FigureNumber',plot,...
              'bgc',bgc,  ...
              'Subfigure',[1 2 2],      ...
              'Color', 'r');
  end
  plotCorrespondences3D([cXr;cXq],...
             'Subfigure',[1 2 2]);
 plotcov3(newPose.H(1:3,4),his.e(iter).covM(1:3,1:3));
 plotPoses3D(HRt2xyzrxryrz(newPose.H));
 plotPoses3D(zeros(6,1));
 cameratoolbar
end

% The new pose is returned according with ICP's updates
newRT = newPose.H;
% newRT = Rt(iter).H;
covM = his.e(iter).covM;
% Means Square Error
Err.dk2 = err(iter,2);
% Covariance from scan matching
% TODO: to compute the covariance according with the minimization used
%Err.covM = naiveCovariance(cXr,cXq);
Err.loop = loop;


%%

if textdisp
    fprintf('Convergence %d :\n ',loop)
end
if nargout == 3 
    varargout(1) = {err};
end

if ~textdisp
    fprintf('\n')
end
