% SAMPLEDATA3D
% Sample 3D clouds according with the indicated fields
%
%% Description
%
% Author - 	
%       Ernesto Homar Teniente Aviles
% 			CONACYT Scholarship
%       Institut de Robotica i Informatica Industrial, CSIC-UPC
%       Last change - 11/2010
%
% This code has been developed as part of the EU URUS Project
% FP6-IST-045062
% 
% Usage
%       samp_data = sampleData(data[,OPTIONS])
%
% Variables
%   Inputs :
%       data  - a 3xn 3 matrix, data (x,y,z)
%               a string, file names to sample
% 
%   Options:
%       'SensorOrigin'    - 3x1 vector, sensor's origin coordinates
%       'UniformSampling'  - +real, bos size
%       'ReductionFilter' -
%       'Median Filter'   -
%       'XYLimit'         -
%       'EraseUpX'        - a real, maximum desired x
%       'EraseUpY'        - a real, maximum desired y
%       'EraseUpZ'        - a real, maximum desired z
%       'EraseUnderX'     - a real, minimum desired x
%       'EraseUnderY'     - a real, minimum desired y
%       'EraseUnderZ'     - a real, minimum desired z
%       'RandomSampleing' - +integer, desired number of points
% 
%   Outputs:
%       samp_data - 3xn matrix, sampled data.
%

%% Code
function samp_data = sampleData3D(data,varargin)

if isa(data, 'string')
  varargin = varargin(1,2:end);
end

chkvarargin(varargin,2);

[ s_origin ,  ...
  box_sz ,    ...
  red_fil ,   ...
  nmed_pts ,  ...
  max_xy ,    ...
  max_x,      ...
  max_y,      ...
  max_z,      ...
  min_x,      ...
  min_y,      ...
  min_z,      ...
  rdn_pts     ...
  max_rad     ...
  min_rad]= process_options(varargin,       ...
                      'SensorOrigin', [0;0;0],      ...
                      'UniformSampling', [], ...
                      'ReductionFilter',[],...
                      'Median Filter',[],    ...
                      'XYLimit',[],       ...
                      'EraseUpX',[],        ...
                      'EraseUpY',[],     ...
                      'EraseUpZ',[],     ...
                      'EraseUnderX',[],     ...
                      'EraseUnderY',[],     ...
                      'EraseUnderZ',[],     ...
                      'RandomSampler',[], ...
                      'MaxRadius', [], ...
                      'MinRadius',[]);



[m n] = size(data);

if (m == 1 && isa(data,'char'))
  %It means a file name is given
  tmp = parseHokuyo3DOneScan(data,'OnlyPoints',1);
  data = tmp.points;
  
%   fprintf('Sampling the data... \n')
  clear tmp
elseif m==3
  % It means a matrix with the points is given
%   fprintf('Sampling the data... \n')
else
  error('Invalida input data, imposible do execute the sampling')
end
[m n] = size(data);

if (~isempty(nmed_pts))
  % Median filter
  data = medianfilter3D(data,nPts);
end
if ~isempty(red_fil)
    if (~isempty(npts_scan2D && npts_scan2D)) 
      % npts_scan2D: size of the 2D laser slice
        k =1;
        tmp_data2 =[];
        for j = 1:npts_scan2D:size(data,1)
            %tmpData = medianfilter3D(N(j).data(i:i+528,:),5);
            if ~mod(k,md)
              tmp_data = data(:,k:k+npts_scan2D-1); 
              tmp_data = reductionfilter3D(tmp_data,red_sz);
              tmp_data2 = [tmp_data2 tmp_data];
            end
            k = k+npts_scan2D;
        end
        data = tmp_data2;
    else
       data = reductionfilter3D(data,red_sz);
    end 
end 
if (~isempty(max_xy)&& max_xy>0)% x and y maximum distance from the given data pose
   data =maxDist(data,s_origin,max_xy);
end

if ~isempty(max_z)
  % Maximum desired z 
   data = eraseUpperPoints3D(data,'z',max_z,'SensorOrigin', s_origin);
end
if ~isempty(min_z) 
  % Minimum desired z 
   data = eraseLowerPoints3D(data,'z',min_z,'SensorOrigin', s_origin);
end

if ~isempty(max_y)
  % Maximum desired y
   data = eraseUpperPoints3D(data,'y',max_y,'SensorOrigin', s_origin);
end
if ~isempty(min_y) 
  % Minimum desired y
   data = eraseLowerPoints3D(data,'y',min_y,'SensorOrigin', s_origin);
end
if ~isempty(max_x)
  % Maximum desired x
   data = eraseUpperPoints3D(data,'x',max_x,'SensorOrigin', s_origin);
end
if ~isempty(min_x) 
  % Minimum desired x
   data = eraseLowerPoints3D(data,'x',min_x,'SensorOrigin', s_origin);
end
if ~isempty(max_rad)
  data =radiusFilter(data,max_rad,min_rad);
elseif ~isempty(min_rad)
  data =radiusFilter(data,max_rad,min_rad);
end 


if  rdn_pts %TODO: change dimensions
  data = randomSamp(data', rdn_pts);
end
    %      plot3(pose(j:j+1,1),pose(j:j+1,2),pose(j:j+1,3),'r.','erasemode','no
%  fprintf('Cloud  Size :  Original %d, sampled : %d \n', [n size(data,2)]);

samp_data = data; 
