% CREATEPTSPLANE -
% Given the plane equation parameters a,b,c and d create a
% points plane
%
%% Description
%
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Scholarship
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 08/2009
% 
% Usage
%       [plane_pts] = createPtsPlane(a,b,c,d[,OPTIONS])
%
% Varibles:
%   Inputs
%       a,b,c,d - integers, plane equation parameters
% 
%   Outputs:
%       plane_pts - a 3 x n matrix, points in the same plane
%
%   Options:
%       'Sampling' - a string, sampling distribution
%                   * 'uniform' or 'random'
%                   * DEFAULT 'random'
%       'SamplingPointsNumber' - a positive integer
%                   * for a random sampling is the total number of point
%                   * for an uniform sampling you will have n^2 points
%       'PlaneInterval' - a 1x3 vector, interval to create plane mesh
%                   * [interval.low interval.up interval.step]
%                   * DEFAULT [-10 10 0.1]
%
%   Note:
%       Plane equation, ax + by + cz + d = 0

%% CODE
function plane_pts = createPtsPlane(a,b,c,d,varargin)

nvarargin = size(varargin,2);
if ~isempty(varargin)
    if mod(nvarargin,2) 
        error('Not enough input arguments, or invalid parameters!!!');
    end
end

% DEFAULTS

sampling.kind = 'random';
sampling.random.n_pts = 1000;
sampling.uniform.n_pts = 25;
interval.low = -10;
interval.up = 10;
interval.step = 0.1;


% READ OPTIONS
if nvarargin >= 2 
    for i=1:2:nvarargin
        % variable = varargin(i);
        switch char(varargin(i)) 
            case 'Sampling'
                sampling = varargin{i+1};
            case 'PlaneInterval' % mesh interval parameters
                tmp = varargin{1,2};
                interval.low =  tmp(1);
                interval.up = tmp(2);
                interval.step = tmp(3);
        end
    end
end

% mesh interval
x=interval.low:interval.step:interval.up;
% (x,y) points
[X,Y] = meshgrid(x);
% corresponding z points
Z=(d-a*X-b*Y)/c;


i = 1;
switch sampling.kind
    case 'uniform'
        p = zeros(3,sampling.uniform.n_pts^2);
        for j=1:sampling.uniform.n_pts
           for k = 1:sampling.uniform.n_pts
              p(:,i) =  [X(j,k),Y(j,k),Z(j,k)];
              i = i+1;
           end
        end
    case 'random'
        n_pts = sampling.random.n_pts;
        p = zeros(3,n_pts);
        n = size(x,2);
        j_indx = ceil(1 + (n-1) * rand(n_pts,1));
        k_indx = ceil(1 + (n-1) * rand(n_pts,1));
%         j_indx = sort(ceil(rand(n_pts,1)*n));
%         k_indx = sort(ceil(rand(n_pts,1)*n));
        for j=1:n_pts
            p(:,i) =  [X(j_indx(j),k_indx(j)),Y(j_indx(j),k_indx(j)),Z(j_indx(j),k_indx(j))];
            i = i+1;

        end
end

plane_pts = p;

