%% findDataNormals
% Given a 3D points set find the normal for each point acording its n 
% nearest neighbors
%
%% Description
% Author - 	
%           Ernesto Homar Teniente Aviles
% 			CONACYT Fellow
%			Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change: 12/2008
%
% This code has been developed as part of the EU URUS Project
% FP6-IST-045062
%
% Modified by:
%           Sergi Foix Salmeron
%           PRE-JAE
%           Institut de Robotica i Informatica Industrial, CSIC-UPC
%           Last change - July, 2009
%
% This code has been developed as part of the XXX Project
% 
%% Usage
%   nnidx =useANN(queryPts,refPts[,OPTIONS])
% Variables
%   Inputs:
%       queryPts 	- a 3 x n matrix, points searching its neighbors
%
%   Options:
%       'maxDist'   - a positive real, maximum distance for filtering ANN, 
%                     DEFAULT = infinit
%       'NNN'       - a positive integer Number of nearest neighbors, 
%                     DEFAULT = 13
%
%   Outputs:
%       normals     - a m x 3 matrix, Normal for each point
%       validPoints - a m x 3 matrix, Valid points
%       errors      - scalar, accumulated distance error of the points that
%                     define the plane with the mean point value.

%% Code
function  [ normals validPoints errors ] = findDataNormals(queryPts, varargin)

narg = length(varargin);

if (mod(narg, 2))
  error('Each option must be a string/value pair.');
end

%% Default values
NNN = 13;
maxDist = 100000;

for ii=1:2:narg
    switch char(varargin(ii))
        case 'NNN'
            NNN = varargin{ii+1};
        case 'maxDist'
            maxDist = varargin{ii+1};
    end
end
clear ii;

if NNN == 1
   error('NNN value most be bigger than 1') 
end

clear narg varargin

%% Calculates the Mean value for each point and its neighbours
[annIndex annDistances] = annquery(queryPts, queryPts, NNN, 'search_sch', 'fr', 'radius', maxDist);
[rr cc] = size(annDistances);
dm = zeros(1,cc);
for ii = 1:cc
   dm(1,ii)= mean(annDistances(2:rr,ii));
end
clear ii;

%% Finds the normal vector that supports the plane of each point neighbours

normals = zeros(3,cc);
validPoints = zeros(1,cc);
errors = zeros(1,cc);

for ii = 1:cc
    % Filter: If the median distance is smaller than a Maxim defined
    % distance then the normal vector is calculated and the related point
    % is defined as valid.
    % If the median is greater the normal is not calculated and the point
    % is defined as not valid.
    if dm(ii) < maxDist
%         points = queryPts(:,annIndex(1:NNN,ii));
        points = queryPts(:,annIndex(find(annIndex(1:NNN,ii)),ii));
        % finds the normal for i and the median point that defines the plane
        [ normal medianPoint relatedError ] = findPointNormal(points, 'type', 'cetto');
        % it constrains the normal vector depending on the camera point of
        % view
        if (acos(dot(-medianPoint, normal)/norm(medianPoint)*norm(normal)) > pi/2) || (acos(dot(-medianPoint, normal)/norm(medianPoint)*norm(normal)) < -pi/2)
            normal = -normal;
        end

        normals(:,ii)= normal;
        errors(ii) = relatedError;
        validPoints(ii) = 1;
    end
end
clear ii;